/* Code to demonstrate lifetime error in ca_{create,clear}_subscription: to be
 * precise, callback occurs after ca_clear_subscription has returned. */

#undef _POSIX_C_SOURCE
#define _POSIX_C_SOURCE 200112L     // Needed for sem_timedwait
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <time.h>
#include <semaphore.h>

#include <cadef.h>
#include <iocsh.h>


struct channel
{
    char name[20];
    chid channel_id;
    sem_t connected;
};

struct event
{
    bool valid;
    evid event_id;
};


static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
#define LOCK()       pthread_mutex_lock(&mutex)
#define UNLOCK()     pthread_mutex_unlock(&mutex)


static void on_update(struct event_handler_args args)
{
    struct event *event = args.usr;
    LOCK();
    bool valid = event->valid;
    UNLOCK();
    if (valid)
    {
        printf(".");
        fflush(stdout);
    }
    else
    {
        printf("whoops!\n");
        exit(2);
    }
}

static void on_ca_connect(struct connection_handler_args args)
{
    struct channel *channel = ca_puser(args.chid);
    sem_post(&channel->connected);
}

static struct channel *create_channels(void)
{
    struct channel *channels = calloc(PV_COUNT, sizeof(struct channel));
    for (int i = 0; i < PV_COUNT; i ++)
    {
        struct channel *channel = &channels[i];
        sprintf(channel->name, "TESTPV%d", i + 1);
        sem_init(&channel->connected, 0, 0);
        ca_create_channel(
            channel->name, on_ca_connect, channel, 0, &channel->channel_id);
    }
    return channels;
}

static bool connect_channels(struct channel *channels)
{
    struct timespec deadline;
    clock_gettime(CLOCK_REALTIME, &deadline);
    deadline.tv_sec += 1;
    for (int i = 0; i < PV_COUNT; i ++)
    {
        struct channel *channel = &channels[i];
        if (sem_timedwait(&channel->connected, &deadline) != 0)
        {
            printf("Unable to connect channel %s\n", channel->name);
            return false;
        }
    }
    printf("All channels connected\n");
    return true;
}

static struct event *create_events(struct channel *channels)
{
    /* Note: new event objects allocated on each cycle rather than being reused
     * so that we can more reliably detect the lifetime error. */
    struct event *events = calloc(PV_COUNT, sizeof(struct event));
    for (int i = 0; i < PV_COUNT; i ++)
    {
        struct channel *channel = &channels[i];
        struct event *event = &events[i];
        LOCK();
        event->valid = true;
        UNLOCK();
        ca_create_subscription(DBR_LONG, 1, channel->channel_id, DBE_VALUE,
            on_update, event, &event->event_id);
    }
    return events;
}

static void close_events(struct event *events)
{
    for (int i = 0; i < PV_COUNT; i ++)
    {
        struct event *event = &events[i];
        LOCK();     // <----(1) Should trigger deadlock, instead helps crash
        ca_clear_subscription(event->event_id);
//         LOCK();     // <----(2) Makes crash more elusive
        event->valid = false;
        UNLOCK();
    }
}

int main(int argc, char **argv)
{
    int cycles = 2;
    if (argc > 1)
        cycles = atoi(argv[1]);
    int interval = 300;
    if (argc > 2)
        interval = atoi(argv[2]);

    ca_context_create(1);
    iocsh("TEST.init");

    struct channel *channels = create_channels();
    if (!connect_channels(channels))
        return 1;

    printf("Testing %d cycles, interval %d us\n", cycles, interval);
    for (int c = 0; c < cycles; c ++)
    {
        printf("["); fflush(stdout);
        struct event *events = create_events(channels);
        ca_flush_io();          // Makes no observable difference
        usleep(interval);
        close_events(events);
        printf("]"); fflush(stdout);
    }

    return 0;
}
